{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OverloadedStrings #-}

module Lisa
    ( app
    , mkConfig
    ) where

import Control.Concurrent (MVar, newMVar, putMVar, takeMVar)
import Control.Monad.IO.Class (MonadIO, liftIO)
import Data.Maybe (fromMaybe)
import Data.Time.Clock (UTCTime, getCurrentTime, addUTCTime)
import Network.HTTP.Req (https)
import Text.Blaze.Html (Html)
import Text.Blaze.Html.Renderer.Utf8 (renderHtml)
import Text.Printf (printf)
import Web.Spock (SpockAction, SpockM, ActionCtxT, get, getState, lazyBytes, param, root, setHeader, var, (<//>))
import Web.Spock.Config (PoolOrConn(PCNoDatabase), SpockCfg, defaultSpockCfg)

import Lisa.Squeak (SqueakCtx, SqueakError, Lecture)

import qualified Lisa.Squeak as Squeak
import qualified Lisa.Views as Views

-- Utilities

-- From https://github.com/Rockfordal/spockmini/blob/51af99760bbf68c956e906c9633182700a18be9c/src/Web/Utils.hs#L17
blaze :: MonadIO m => Html -> ActionCtxT ctx m a
blaze html = do
    setHeader "Content-Type" "text/html; charset=utf-8"
    lazyBytes $ renderHtml html

instance SqueakCtx (SpockAction () () State) where
    getLocationInfo = pure $ Right (https "api.squeak-test.fsmi.uni-karlsruhe.de", mempty)
    getAuthToken = pure Nothing

-- State

data State = State
    { stateLectures :: MVar (UTCTime, [Lecture]) -- ^ Cached lectures with cache expiration time.
    }

-- | Get lectures from cache if they've been retrieved before or from Squeak.
-- Uses 'MVar's for synchronisation: As long as each thread 'takeMVar's before it 'putMVar's, we're in the clear.
-- TODO: Evaluate lectures to NF before 'putMVar'ing them.
getAllLecturesCached :: SpockAction () () State (Either SqueakError [Lecture])
getAllLecturesCached = do
    lecturesMVar <- stateLectures <$> getState
    (expirationTime, lectures) <- liftIO $ takeMVar lecturesMVar
    now <- liftIO getCurrentTime
    (newContents, result) <- if now >= expirationTime
    then -- Cache expired => fetch lectures again.
        Squeak.getAllLectures >>= \case
            -- On error: Clear cache.
            Left err ->
                pure ((now, []), Left err)
            -- On success: Cache fetched lectures for 60 seconds.
            Right lectures' -> do
                let expirationTime' = addUTCTime 60 now
                liftIO $ printf "Fetched %d lectures, valid until %s\n" (length lectures') (show expirationTime')
                pure ((expirationTime', lectures'), Right lectures')
    else -- Cache still valid => return it.
        pure ((expirationTime, lectures), Right lectures)
    liftIO $ putMVar lecturesMVar newContents
    pure result

-- Exports

mkConfig :: IO (SpockCfg () () State)
mkConfig = do
    now <- getCurrentTime
    lecturesMVar <- newMVar (now, [])
    defaultSpockCfg () PCNoDatabase (State lecturesMVar)

app :: SpockM () () State ()
app = do
    get root $ do
        blaze Views.viewIndex
    
    get ("lectures" <//> var) $ \id -> do
        documents <- Squeak.getDocumentsByLectureId id
        blaze $ either Views.viewSqueakError (Views.viewLecture id) documents

    get "lectures" $ do
        query <- fromMaybe "" <$> param "query"
        lectures <- getAllLecturesCached
        blaze $ either Views.viewSqueakError (Views.viewLectures query) lectures
